from io import BytesIO

from pytest import fixture

from phylodata.process.evolutionary_model.parse_beast2_packages import (
    BEAST2_PACKAGE_PARSERS,
    BEAST2_PACKAGE_PARSERS_PER_NAME,
    AutoGeneratedBeast2PackageParser,
)
from phylodata.process.evolutionary_model.parse_evolutionary_model import (
    parse_evolutionary_model,
)


def to_bytes_io(text: str):
    bytesio = BytesIO(bytes(text, "ascii"))
    bytesio.name = "Test"
    return bytesio


@fixture
def babel_model():
    package = AutoGeneratedBeast2PackageParser("Babel", "treePrior", "babel")
    BEAST2_PACKAGE_PARSERS.clear()
    BEAST2_PACKAGE_PARSERS.append(package)
    BEAST2_PACKAGE_PARSERS_PER_NAME["Babel"] = package


def test_detect_babel_in_specs_with_attributes(babel_model):
    file = to_bytes_io("""<?xml version="1.0" encoding="UTF-8" standalone="no"?>
    <beast namespace="beast.core" required="BEAST.base v2.7.7" version="2.7">
        <data></data>
        <run>
            <distribution
                    id="treeLikelihood.ringe"
                    spec="babel.evolution.likelihood.ALSTreeLikelihood"
                    tree="@Tree.t:ringe"
                    useAmbiguities="true"
                >
                <observationprocess
                        id="AnyTipObservationProcess.0"
                        spec="babel.evolution.likelihood.AnyTipObservationProcess"
                        integrateGainRate="true"
                        mu="@deathRate.SDollo"
                        tree="@Tree.t:ringe"
                    >
                </observationprocess>
            </distribution>
        </run>
    </beast>""")
    model = parse_evolutionary_model(file)

    assert model
    assert len(model) == 1
    assert model[0].name == "Babel"


def test_detect_babel_in_specs_with_tags(babel_model):
    file = to_bytes_io("""<?xml version="1.0" encoding="UTF-8" standalone="no"?>
    <beast namespace="beast.core" required="BEAST.base v2.7.7" version="2.7">
        <data></data>
        <run>
            <distribution
                    id="treeLikelihood.ringe"
                    tree="@Tree.t:ringe"
                    useAmbiguities="true"
                >
                <spec>babel.evolution.likelihood.ALSTreeLikelihood</spec>
                <observationprocess
                        id="AnyTipObservationProcess.0"
                        spec="babel.evolution.likelihood.AnyTipObservationProcess"
                        integrateGainRate="true"
                        mu="@deathRate.SDollo"
                        tree="@Tree.t:ringe"
                    >
                </observationprocess>
            </distribution>
        </run>
    </beast>""")
    model = parse_evolutionary_model(file)

    assert model
    assert len(model) == 1
    assert model[0].name == "Babel"


def test_detect_babel_in_specs_with_input_tag(babel_model):
    file = to_bytes_io("""<?xml version="1.0" encoding="UTF-8" standalone="no"?>
    <beast namespace="beast.core" required="BEAST.base v2.7.7" version="2.7">
        <data></data>
        <run>
            <distribution
                    id="treeLikelihood.ringe"
                    tree="@Tree.t:ringe"
                    useAmbiguities="true"
                >
                <input name="spec">babel.evolution.likelihood.ALSTreeLikelihood</input>
                <observationprocess
                        id="AnyTipObservationProcess.0"
                        spec="babel.evolution.likelihood.AnyTipObservationProcess"
                        integrateGainRate="true"
                        mu="@deathRate.SDollo"
                        tree="@Tree.t:ringe"
                    >
                </observationprocess>
            </distribution>
        </run>
    </beast>""")
    model = parse_evolutionary_model(file)

    assert model
    assert len(model) == 1
    assert model[0].name == "Babel"


def test_detect_babel_in_namespace(babel_model):
    file = to_bytes_io("""<?xml version="1.0" encoding="UTF-8" standalone="no"?>
    <beast namespace="beast.core:babel.evolution.likelihood" required="BEAST.base v2.7.7" version="2.7">
        <data></data>
        <run>
            <distribution
                    id="treeLikelihood.ringe"
                    spec="ALSTreeLikelihood"
                    tree="@Tree.t:ringe"
                    useAmbiguities="true"
                >
                <observationprocess
                        id="AnyTipObservationProcess.0"
                        spec="AnyTipObservationProcess"
                        integrateGainRate="true"
                        mu="@deathRate.SDollo"
                        tree="@Tree.t:ringe"
                    >
                </observationprocess>
            </distribution>
        </run>
    </beast>""")
    model = parse_evolutionary_model(file)

    assert model
    assert len(model) == 1
    assert model[0].name == "Babel"


def test_detect_babel_in_map(babel_model):
    file = to_bytes_io("""<?xml version="1.0" encoding="UTF-8" standalone="no"?>
    <beast namespace="beast.core" required="BEAST.base v2.7.7" version="2.7">
        <data></data>
        <map value="B">babel.evolution</map>
        <run>
            <distribution
                    id="treeLikelihood.ringe"
                    spec="ALSTreeLikelihood"
                    tree="@Tree.t:ringe"
                    useAmbiguities="true"
                >
                <observationprocess
                        id="AnyTipObservationProcess.0"
                        spec="AnyTipObservationProcess"
                        integrateGainRate="true"
                        mu="@deathRate.SDollo"
                        tree="@Tree.t:ringe"
                    >
                </observationprocess>
            </distribution>
        </run>
    </beast>""")
    model = parse_evolutionary_model(file)

    assert model
    assert len(model) == 1
    assert model[0].name == "Babel"



from io import BytesIO

import streamlit as st

from phylodata.data_types import EvolutionaryModelComponent
from phylodata.process.evolutionary_model.parse_evolutionary_model import (
    parse_evolutionary_model,
)
from phylodata.process.module import Module


class EvolutionaryModelModule(Module[list[EvolutionaryModelComponent]]):
    def ui(self): ...

    def validate(self): ...

    def parse(self, beast2_configuration: BytesIO) -> list[EvolutionaryModelComponent]:
        st.text("Detecting the evolutionary model...")
        return parse_evolutionary_model(beast2_configuration)



import re

import nanoid


def generate_human_readable_id(title: str, year: int, authors: list[str]) -> str:
    """Generates a human-readable ID for the paper. Multiple calls with the same
    metadata will generate different IDs, as a random suffix is added.

    It is designed to be both recognizable and sufficiently random to avoid collisions.

    It generates IDs like 'felsenstein-1973-maximum-dsa6'."""
    author = replace_non_alpha(authors[0]).split()[-1].lower()
    title_excerpt = next(
        (
            word.lower()
            for word in replace_non_alpha(title).split()
            if word.lower() not in STOP_WORDS
        ),
        "",
    )
    random_id = nanoid.generate(size=4, alphabet="1234567890abcdefghijklmnopqrstuvwxyz")
    return f"{author}-{year}-{title_excerpt}-{random_id}"


def replace_non_alpha(text: str) -> str:
    """Replaces all non-alphabetical characters with a blank."""
    return re.sub("[^0-9a-zA-Z]+", " ", text)


STOP_WORDS = [
    "i",
    "me",
    "my",
    "myself",
    "we",
    "our",
    "ours",
    "ourselves",
    "you",
    "your",
    "yours",
    "yourself",
    "yourselves",
    "he",
    "him",
    "his",
    "himself",
    "she",
    "her",
    "hers",
    "herself",
    "it",
    "its",
    "itself",
    "they",
    "them",
    "their",
    "theirs",
    "themselves",
    "what",
    "which",
    "who",
    "whom",
    "this",
    "that",
    "these",
    "those",
    "am",
    "is",
    "are",
    "was",
    "were",
    "be",
    "been",
    "being",
    "have",
    "has",
    "had",
    "having",
    "do",
    "does",
    "did",
    "doing",
    "a",
    "an",
    "the",
    "and",
    "but",
    "if",
    "or",
    "because",
    "as",
    "until",
    "while",
    "of",
    "at",
    "by",
    "for",
    "with",
    "about",
    "against",
    "between",
    "into",
    "through",
    "during",
    "before",
    "after",
    "above",
    "below",
    "to",
    "from",
    "up",
    "down",
    "in",
    "out",
    "on",
    "off",
    "over",
    "under",
    "again",
    "further",
    "then",
    "once",
    "here",
    "there",
    "when",
    "where",
    "why",
    "how",
    "all",
    "any",
    "both",
    "each",
    "few",
    "more",
    "most",
    "other",
    "some",
    "such",
    "no",
    "nor",
    "not",
    "only",
    "own",
    "same",
    "so",
    "than",
    "too",
    "very",
    "s",
    "t",
    "can",
    "will",
    "just",
    "don",
    "should",
    "now",
]
"""The list of stop words taken from NLTK."""



import os
import shutil
from io import BytesIO
from pathlib import Path
from typing import Optional

import msgspec

from phylodata.data_types import (
    EditableExperiment,
    EditablePaper,
    EditablePaperWithExperiment,
    EvolutionaryModelComponent,
    File,
    FileType,
    Metadata,
    NonEditableExperiment,
    NonEditablePaper,
    NonEditablePaperWithExperiment,
    Sample,
    Trees,
)
from phylodata.process.utils.file_utils import add_file_name_suffix

WASABI_BUCKET_NAME = "phylodata-experiments"

EDITABLE_METADATA_FILE = "editable_phylodata_metadata.json"
NON_EDITABLE_METADATA_FILE = "non_editable_phylodata_metadata"


def generate_output(
    editable_experiment: EditableExperiment,
    non_editable_experiment: NonEditableExperiment,
    editable_paper: EditablePaper,
    non_editable_paper: NonEditablePaper,
    samples: list[Sample],
    files: list[File],
    evolutionary_model: list[EvolutionaryModelComponent],
    trees: Optional[Trees],
    metadata: Metadata,
) -> Path:
    """Creates the output folder for the given experiment metadata."""
    files = clean_up_file_names(files)
    files = add_non_editable_metadata_file(
        files,
        non_editable_paper,
        non_editable_experiment,
        evolutionary_model,
        trees,
        metadata,
    )
    files = add_editable_metadata_file(
        files, editable_paper, editable_experiment, samples
    )

    output_folder = create_output_folder(non_editable_experiment)
    store_files(files, output_folder)

    return output_folder


def clean_up_file_names(files: list[File]) -> list[File]:
    """Makes sure all file names are unique and do not correspond to the
    metadata file names."""
    filename_counter = {EDITABLE_METADATA_FILE: 1, NON_EDITABLE_METADATA_FILE: 1}

    for file in files:
        if file.name in filename_counter:
            filename_counter[file.name] += 1
            file.name = add_file_name_suffix(
                file.name, str(filename_counter[file.name]), "_"
            )
        else:
            filename_counter[file.name] = 1

    return files


def add_editable_metadata_file(
    files: list[File],
    editable_paper: EditablePaper,
    editable_experiment: EditableExperiment,
    samples: list[Sample],
) -> list[File]:
    """Adds a new File object to the files list corresponding to the
    editable metadata json."""
    editable_metadata = EditablePaperWithExperiment(
        editable_paper, editable_experiment, samples
    )

    data_bytes = msgspec.json.format(msgspec.json.encode(editable_metadata), indent=2)
    data_bytes_io = BytesIO()
    data_bytes_io.write(data_bytes)

    editable_metadata_file = File.from_bytes(
        data_bytes_io,
        name=EDITABLE_METADATA_FILE,
        type=FileType.PHYLO_DATA_EXPERIMENT,
    )
    files.append(editable_metadata_file)

    return files


def add_non_editable_metadata_file(
    files: list[File],
    non_editable_paper: NonEditablePaper,
    non_editable_experiment: NonEditableExperiment,
    evolutionary_model: list[EvolutionaryModelComponent],
    trees: Optional[Trees],
    metadata: Metadata,
) -> list[File]:
    """Adds a new File object to the files list corresponding to the
    non-editable metadata json."""
    non_editable_metadata = NonEditablePaperWithExperiment(
        files=files,
        experiment=non_editable_experiment,
        paper=non_editable_paper,
        evolutionary_model=evolutionary_model,
        trees=trees,
        metadata=metadata,
    )

    data_bytes = msgspec.json.encode(non_editable_metadata)
    data_bytes_io = BytesIO()
    data_bytes_io.write(data_bytes)

    non_editable_metadata_file = File.from_bytes(
        data_bytes_io,
        name=NON_EDITABLE_METADATA_FILE,
        type=FileType.PHYLO_DATA_EXPERIMENT,
    )
    files.append(non_editable_metadata_file)

    return files


def create_output_folder(experiment: NonEditableExperiment) -> Path:
    """Creates the output folder. If the folder already exists,
    it gets deleted first."""
    output_folder = Path(experiment.human_readable_id)

    if os.path.exists(output_folder):
        shutil.rmtree(output_folder)
    os.mkdir(output_folder)

    return output_folder


def store_files(files: list[File], output_folder: Path):
    """Stores the files in the output folder."""
    for file in files:
        if not file.bytes:
            raise ValueError(
                f"Trying to store a file without bytes ({file.name}). This should not happen."
            )

        with open(f"{output_folder}/{file.name}", "wb") as f:
            f.write(file.bytes.getbuffer())



 phylodata

"""
Download single experiment without using it at all in python.
"""
phylodata.load_experiment(
    "gavryushkina-2017-bayesian", directory="data", only_preview=False
)

"""
Download single experiment, but only certain files that satisify a grep pattern.
"""
phylodata.load_experiment(
    "gavryushkina-2017-bayesian", directory="data", only_preview=False, files="*khk.png"
)

"""
Download multiple experiments without using it at all in python.
"""
phylodata.load_experiments(
    ["gavryushkina-2017-bayesian/1", "drummond-2010-bayesian/1"],
    directory="data",
    only_preview=False,
)

"""
Download single experiment and print all files.
"""
experiment = phylodata.load_experiment("gavryushkina-2017-bayesian", directory="data")

for file in experiment:
    print(file.name)

"""
Download single experiment and load the trees.
"""
experiment = phylodata.load_experiment("gavryushkina-2017-bayesian", directory="data")

for file in experiment:
    if file.type == FileType.BEAST2_POSTERIOR_TREES:
        trees_path = file.path
        break

trees_path = experiment[FileType.BEAST2_POSTERIOR_TREES].path

trees_path = experiment.files.posterior_trees
trees_path = experiment.posterior_trees
trees_path = get_posterior_trees_file(experiment)

"""
Download single experiment and load the samples.
"""
experiment = phylodata.load_experiment("gavryushkina-2017-bayesian", directory="data")

for sample in experiment.samples:
    print(sample.scientific_name)

"""
Only download metadata.
"""
experiment = phylodata.load_experiment(
    "gavryushkina-2017-bayesian", files_to_download=[]
)

for file in experiment:
    print(file.name)


"""
Load experiment from folder.
"""
experiment = phylodata.load_experiment(directory="jlsafd/gavryushkina-2017-bayesian")

for file in experiment:
    print(file.name)


"""
Download single experiment and load the trees using ecosystem library.
"""
import phylotrees

experiment = phylodata.load_experiment(
    "gavryushkina-2017-bayesian", directory="data", version=10
)
trees = phylotrees.load_trees(experiment)



"""Generate JSON schema for EditablePaperWithExperiment data type."""

import msgspec

from phylodata.data_types import (
    EditablePaperWithExperiment,
    NonEditablePaperWithExperiment,
)

editable_schema = msgspec.json.schema(EditablePaperWithExperiment)
with open("../website/src/lib/schema/editableSchema.json", "wb") as f:
    f.write(msgspec.json.format(msgspec.json.encode(editable_schema)))

non_editable_schema = msgspec.json.schema(NonEditablePaperWithExperiment)
with open("../website/src/lib/schema/nonEditableSchema.json", "wb") as f:
    f.write(msgspec.json.format(msgspec.json.encode(non_editable_schema)))



import csv
import re
from pathlib import Path
from typing import TypedDict

import phylodata


def clean_label(label: str) -> str:
    label = label.replace("sequence", "")
    label = label.replace("seq", "")
    return re.sub(r"[^A-Za-z ]+", " ", label).lower().strip()


class Language(TypedDict):
    language_id: str

    parent_id: str
    name: str


languages: list[Language] = []
id_to_language: dict[str, Language] = {}
cleaned_language_names: list[str] = []

# the following is a hacky way to retrieve the glottolog languoid data file
# TODO: check if there is a better way to do this
with open(
    Path(phylodata.__path__[0]) / "process" / "data" / "glottolog_languoid_5.2.csv"  # type: ignore
) as csv_file:  # type: ignore
    glottolog_languages = csv.reader(csv_file)

    for row in glottolog_languages:
        languages.append(
            {
                "language_id": row[0],
                "parent_id": row[2],
                "name": row[3],
            }
        )
        id_to_language[row[0]] = {
            "language_id": row[0],
            "parent_id": row[2],
            "name": row[3],
        }
        cleaned_language_names.append(clean_label(row[3]))



def contains_sequence(sequence: str):
    return any([ch for ch in sequence if ch.isalpha()])


def is_amino_acid_sequence(sequence: str):
    characters = {c.lower() for c in sequence}
    return characters.issubset(AA_CHARACTERS)


def is_dna_sequence(sequence: str):
    characters = {c.lower() for c in sequence}
    return characters.issubset(DNA_CHARACTERS)


def is_rna_sequence(sequence: str):
    characters = {c.lower() for c in sequence}
    return characters.issubset(RNA_CHARACTERS)


DNA_CHARACTERS = {
    "a",
    "t",
    "c",
    "g",
    "*",
    "?",
    "-",
    "r",
    "y",
    "s",
    "w",
    "k",
    "m",
    "b",
    "d",
    "h",
    "v",
    "n",
}
RNA_CHARACTERS = (DNA_CHARACTERS - {"t"}) | {"u"}
AA_CHARACTERS = {
    "a",
    "r",
    "n",
    "d",
    "c",
    "e",
    "q",
    "g",
    "h",
    "i",
    "l",
    "k",
    "m",
    "f",
    "p",
    "s",
    "t",
    "w",
    "y",
    "v",
    "u",
    "o",
    "?",
    "*",
    "-",
}



from io import BytesIO

import pytest

from phylodata.errors import ValidationError
from phylodata.process.files.parse_files import parse_beast2_logs


def to_bytes_io(text: str):
    bytesio = BytesIO(bytes(text, "ascii"))
    bytesio.name = "Test"
    return bytesio


def test_simple_log_is_ok():
    raw_logs = "state   posterior"
    for s in range(1000):
        raw_logs += f"\n{s}   0"

    file = to_bytes_io(raw_logs)
    list(parse_beast2_logs(file))


def test_log_file_with_few_lines_fails():
    with pytest.raises(ValidationError):
        raw_logs = "state   posterior"
        for s in range(5):
            raw_logs += f"\n{s} 0"

        file = to_bytes_io(raw_logs)
        list(parse_beast2_logs(file))


def test_empty_file_fails():
    with pytest.raises(ValidationError):
        file = to_bytes_io("")
        list(parse_beast2_logs(file))


def test_no_tsv_fails():
    with pytest.raises(ValidationError):
        file = to_bytes_io("""This is some random file:khklhdsf
            asdfsadfsadf""")
        list(parse_beast2_logs(file))


def test_preview_file_is_returned():
    raw_logs = "state   posterior"
    for s in range(1000):
        raw_logs += f"\n{s}   0"

    file = to_bytes_io(raw_logs)
    assert len(list(parse_beast2_logs(file))) == 2


def test_size_of_preview_file():
    raw_logs = "state   posterior"
    for s in range(1000):
        raw_logs += f"\n{s}   0"

    file = to_bytes_io(raw_logs)
    files = parse_beast2_logs(file)

    full_file = next(files)
    preview_file = next(files)

    assert full_file.bytes and preview_file.bytes
    assert (
        0 < preview_file.bytes.getbuffer().nbytes < full_file.bytes.getbuffer().nbytes
    )



from io import BytesIO

import streamlit as st

from phylodata.data_types import Trees
from phylodata.process.module import Module
from phylodata.process.trees.parse_trees import parse_trees


class TreesModule(Module[Trees | None]):
    def ui(self): ...

    def validate(self): ...

    def parse(self, beast2_trees: BytesIO | None) -> Trees | None:
        if not beast2_trees:
            return None
        st.text("Processing the trees...")
        return parse_trees(beast2_trees)



from io import BytesIO

import streamlit as st

from phylodata.data_types import Sample
from phylodata.process.module import Module
from phylodata.process.samples.parse_beast2_samples import parse_beast2_samples


class SamplesModule(Module[list[Sample]]):
    def ui(self): ...

    def validate(self): ...

    def parse(self, beast2_configuration: BytesIO) -> list[Sample]:
        st.text(
            "Detecting the samples (e.g. using BEAST2 or language lookup). "
            "This might take a while, especially for amino acid sequences..."
        )
        return parse_beast2_samples(beast2_configuration)



import json
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any
from xml.etree.ElementTree import ElementTree

import phylodata
from phylodata.data_types import ModelType
from phylodata.process.utils.beast2_xml_utils import get_attribute


class Beast2PackageParser(ABC):
    """An abstract base class for parsing BEAST 2 packages."""

    def is_used(self, beast2_xml: ElementTree) -> bool:
        """Returns true if the package is detected in the given BEAST2 xml.
        Checks the top-level namespaces and the spec attribute of all elements."""
        root = beast2_xml.getroot()
        if root is None:
            return False

        # check namespaces

        namespaces = get_attribute(root, "namespace", "")
        namespaces = namespaces.split(":")

        package_namespace = self.get_namespace()

        if namespaces and any(
            namespace == package_namespace
            or namespace.startswith(f"{package_namespace}.")
            for namespace in namespaces
        ):
            return True

        # check maps

        elements_with_map = root.findall(".//map")
        if any(
            elem.text and elem.text.startswith(package_namespace)
            for elem in elements_with_map
        ):
            return True

        # check specs

        elements_with_spec = root.findall(".//*[@spec]")
        spec_values = [get_attribute(elem, "spec") for elem in elements_with_spec]

        for spec in spec_values:
            if not spec:
                continue

            # for the following cases: assume package_namespace = "babel.distribution"

            if spec == package_namespace:
                return True

            if spec.startswith(f"{package_namespace}."):
                # e.g. spec = "babel.distribution.Normal"
                return True

            if any(
                f"{namespace}.{spec}" == package_namespace for namespace in namespaces
            ):
                # e.g. namespace = "babel" and spec = "distribution"
                return True

            if any(
                f"{namespace}.{spec}".startswith(f"{package_namespace}.")
                for namespace in namespaces
            ):
                # e.g. namespace = "babel" and spec = "distribution.Normal"
                return True

        return False

    @abstractmethod
    def get_name(self) -> str:
        """Returns a human-readable name of the BEAST 2 package."""
        raise NotImplementedError

    @abstractmethod
    def get_type(self) -> ModelType:
        """Returns the type of the BEAST 2 package."""
        raise NotImplementedError

    @abstractmethod
    def get_namespace(self) -> str:
        """Returns the namespace of the BEAST 2 package."""
        raise NotImplementedError

    def get_parameters(self, beast2_xml: ElementTree) -> dict[str, Any]:
        """Extracts potential package parameters from the BEAST 2 xml configuration."""
        return {}


class AutoGeneratedBeast2PackageParser(Beast2PackageParser):
    def __init__(self, name: str, type: str, namespace: str):
        self.name = name
        self.type = type
        self.namespace = namespace

    def get_name(self) -> str:
        return self.name

    def get_type(self) -> ModelType:
        return next(t for t in ModelType if t.value == self.type)

    def get_namespace(self) -> str:
        return self.namespace


def load_package_parsers_from_json() -> list[Beast2PackageParser]:
    """Dynamically loads BEAST 2 package parsers using the models JSON file.
    This is useful to sync the models detected by the PhyloData python library
    with the models presented on the PhyloData website."""

    with open(
        Path(phylodata.__path__[0]) / "process" / "evolutionary_model" / "models.json",
        "r",
    ) as handle:  # type: ignore
        json_models = json.loads(handle.read())

        return [
            AutoGeneratedBeast2PackageParser(
                name=model["name"],
                type=model["model"],
                namespace=model["namespaces"][0],
            )
            for model in json_models
        ]

        return []


BEAST2_PACKAGE_PARSERS: list[Beast2PackageParser] = load_package_parsers_from_json()
BEAST2_PACKAGE_PARSERS_PER_NAME: dict[str, Beast2PackageParser] = {
    parser.get_name(): parser for parser in BEAST2_PACKAGE_PARSERS
}



from loguru import logger

from phylodata.data_types import ClassificationEntry, DataType, Sample
from phylodata.errors import BlastError
from phylodata.process.samples.run_blast import extract_taxon_ids, run_blast
from phylodata.process.samples.taxon_classification import look_up_taxon_classification

MAX_SEQ_LENGTH_CONSIDERED = 2500


def add_nucleotide_metadata(samples: list[Sample]) -> list[Sample]:
    """Tries to add metadata to all samples for which we have nucleotide
    sequences."""
    if not samples:
        return []

    max_sequences_per_sample = max([len(sample.sample_data) for sample in samples])

    # we try the first data point for each sample and see if we can get a classification
    # and continue with the other data points until we find one
    for i in range(max_sequences_per_sample):
        nucleotide_sequences: list[str] = []
        nucleotide_sequence_idx: list[int] = []

        for sample_idx, sample in enumerate(samples):
            if sample.classification:
                # we already have found a classification for this sample
                continue

            if len(sample.sample_data) <= i:
                continue

            data = sample.sample_data[i]
            if data.type in [DataType.RNA, DataType.DNA]:
                nucleotide_sequences.append(
                    "".join(c for c in data.data if c.isalpha())
                )
                nucleotide_sequence_idx.append(sample_idx)

        # fetch the nucleotide classifications

        classification = fetch_nucleotide_classifications(nucleotide_sequences)

        # add the classifications to the samples

        for sample_idx, classification in zip(nucleotide_sequence_idx, classification):
            if classification:
                samples[sample_idx].classification = classification
                samples[sample_idx].scientific_name = classification[0].scientific_name
                samples[sample_idx].common_name = classification[0].common_name

    return samples


def fetch_nucleotide_classifications(
    sequences: list[str],
) -> list[list[ClassificationEntry] | None]:
    if not sequences:
        return []

    blast_params = {
        "CMD": "Put",
        "PROGRAM": "blastn",
        "MEGABLAST": True,
        "DATABASE": "core_nt",
        "HITLIST_SIZE": 1,
    }

    try:
        blast_results = run_blast(sequences, blast_params, MAX_SEQ_LENGTH_CONSIDERED)
    except BlastError:
        return [None] * len(sequences)

    taxon_ids = extract_taxon_ids(blast_results)

    classifications: list[list[ClassificationEntry] | None] = []

    for taxon_id in taxon_ids:
        if taxon_id:
            try:
                classifications.append(look_up_taxon_classification(taxon_id))
            except Exception as e:
                logger.error(f"Taxon look up faild: {e}")
                classifications.append(None)
        else:
            classifications.append(None)

    return classifications



import re
from typing import Generator

import bibtexparser


def is_valid_bibtex(bibtex_str: str):
    """Checks if the given string contains at least one valid bibtex entry."""
    return len(bibtexparser.parse_string(bibtex_str).entries) == 1


def get_bibtex_value(bibtex_str: str, key: str) -> str | None:
    """Returns the value for the given key of the first bibtex entry found.
    Returns None if no entry is present or the key does not exist in the first entry."""
    if not (parsed_bibtex := bibtexparser.parse_string(bibtex_str).entries):
        return

    if not (
        attribute := next(
            (
                entry
                for k, entry in parsed_bibtex[0].fields_dict.items()
                if k.lower() == key.lower()
            ),
            None,
        )
    ):
        return

    return attribute.value


def yield_bibtex_authors(authors: str) -> Generator[str, None, None]:
    """Yields the authors in the given bibtex author string. The names of the authors
    are translated to a <First Name> <Last Name> <Suffix>.

    See https://bibtex.eu/fields/author/
    """
    author_list = re.split(" and | AND ", authors) or []
    for author in author_list:
        if "," in author:
            author_parts = author.split(",")

            if len(author_parts) == 2:
                # it's Lastname, Firstname
                author = author_parts[1].strip() + " " + author_parts[0].strip()
            elif len(author_parts) == 3:
                # it's Lastname, Suffix, Firstname
                author = (
                    author_parts[2].strip()
                    + " "
                    + author_parts[0].strip()
                    + author_parts[1].strip()
                )

        yield author



from io import BytesIO

import pytest

from phylodata.errors import ValidationError
from phylodata.process.files.parse_files import parse_beast2_trees


def to_bytes_io(text: str):
    bytesio = BytesIO(bytes(text, "ascii"))
    bytesio.name = "Test"
    return bytesio


def test_simple_nexus_is_ok():
    raw_nexus = """
    #NEXUS
    Begin taxa;
       	Dimensions ntax=3;
        Taxlabels
    taxa_0
    taxa_1
            taxa_2;
    End;
    Begin trees;"""
    for s in range(1000):
        raw_nexus += f"\ntree STATE_{s} = ((0, 1), 2);"
    raw_nexus += "\nEnd;"

    file = to_bytes_io(raw_nexus)
    list(parse_beast2_trees(file))


def test_trees_file_with_few_lines_fails():
    with pytest.raises(ValidationError):
        raw_nexus = """
        #NEXUS
        Begin taxa;
        	Dimensions ntax=3;
        		Taxlabels
        			0
        			1
                    2;
        End;
        Begin trees;
            tree STATE_1 = ((0, 1), 2);
            tree STATE_2 = ((0, 1), 2);
            tree STATE_3 = ((0, 1), 2);
        End;
        """

        file = to_bytes_io(raw_nexus)
        list(parse_beast2_trees(file))


def test_empty_file_fails():
    with pytest.raises(ValidationError):
        file = to_bytes_io("")
        list(parse_beast2_trees(file))


def test_no_nexus_fails():
    with pytest.raises(ValidationError):
        file = to_bytes_io("""This is some random file:khklhdsf
            asdfsadfsadf""")
        list(parse_beast2_trees(file))



from io import BytesIO

from phylodata.data_types import EvolutionaryModelComponent
from phylodata.process.evolutionary_model.parse_beast2_packages import (
    BEAST2_PACKAGE_PARSERS,
)
from phylodata.process.utils.bytesio_utils import get_xml_from_bytesio


def parse_evolutionary_model(
    beast2_config: BytesIO,
) -> list[EvolutionaryModelComponent]:
    xml = get_xml_from_bytesio(beast2_config)

    models = []

    for parser in BEAST2_PACKAGE_PARSERS:
        if parser.is_used(xml):
            model = EvolutionaryModelComponent(
                name=parser.get_name(),
                parameters=parser.get_parameters(xml),
                type=parser.get_type(),
            )
            models.append(model)

    return models



from phylodata.data_types import FileType  # noqa: F401
from phylodata.loader.load_experiment import (  # noqa: F401
    ExperimentToLoad,
    load_experiment,
    load_experiment_from_local_dir,
    load_experiments,
)



import sys
from io import BytesIO
from pathlib import Path

import streamlit as st

from phylodata.data_types import File, FileType
from phylodata.errors import ValidationError
from phylodata.process.files.parse_files import parse_file
from phylodata.process.module import Module


class FilesModule(Module[list[File]]):
    def __init__(self):
        files_given_as_args = [
            path for file in sys.argv[1:] if (path := Path(file)).is_file()
        ]

        self.beast2_configuration = None
        self.beast2_logs = None
        self.beast2_trees = None
        self.other_files = []

        for file in files_given_as_args:
            bytes = BytesIO(file.read_bytes())
            bytes.name = file.name

            if not self.beast2_configuration and file.suffix == ".xml":
                self.beast2_configuration = bytes
            elif not self.beast2_logs and file.suffix == ".log":
                self.beast2_logs = bytes
            elif not self.beast2_trees and file.suffix == ".trees":
                self.beast2_trees = bytes
            else:
                self.other_files.append(bytes)

    def ui(self):
        with st.container(border=True):
            st.write("""
            ### Files
            """)

            if not self.beast2_configuration:
                st.write("""
                    Select the BEAST 2 XML configuration:
                """)

                self.beast2_configuration = st.file_uploader(
                    "BEAST 2 XML configuration", type=["xml"]
                )

            if not self.beast2_logs:
                st.write("""
                If possible, select the file containing the logs (.log):
                """)
                self.beast2_logs = st.file_uploader("BEAST 2 log file", type=["log"])

            if not self.beast2_trees:
                st.write("""
                If possible, select the file containing the posterior trees (.trees):
                """)
                self.beast2_trees = st.file_uploader(
                    "BEAST 2 posterior trees", type=["trees"]
                )

            if not self.other_files:
                st.write("""
                Optionally, you can also select other files (like summary trees) which you consider an important part of your analysis.
                """)

                self.other_files = st.file_uploader(
                    "Other files (optional)", accept_multiple_files=True
                )

    def validate(self):
        if not self.beast2_configuration:
            raise ValidationError("Specify a BEAST 2 configuration file.")

    def parse(self) -> list[File]:
        st.text("Processing the given files...")
        if not self.beast2_configuration:
            raise ValidationError("Specify a BEAST 2 configuration file.")

        files = list(
            parse_file(self.beast2_configuration, FileType.BEAST2_CONFIGURATION)
        )

        if self.beast2_logs:
            files += parse_file(self.beast2_logs, FileType.BEAST2_POSTERIOR_LOGS)

        if self.beast2_trees:
            files += parse_file(self.beast2_trees, FileType.POSTERIOR_TREES)

        for other_file in self.other_files:
            files += parse_file(other_file, FileType.UNKNOWN)

        return files



from rapidfuzz import fuzz, process

from phylodata.data_types import (
    ClassificationEntry,
    ClassificationEntryType,
    DataType,
    Sample,
)
from phylodata.process.samples.language_utils import (
    clean_label,
    cleaned_language_names,
    id_to_language,
    languages,
)


def add_language_metadata(samples: list[Sample]) -> list[Sample]:
    """Tries to add metadata to all samples which correspond to a language."""
    for sample in samples:
        for data in sample.sample_data:
            if data.type not in [DataType.TRAITS, DataType.UNKNOWN]:
                continue

            if sample.classification:
                continue

            if classification := fetch_language_classification(sample.sample_id):
                sample.classification = classification
                sample.scientific_name = classification[0].scientific_name
                break  # we only need one match

    return samples


def fetch_language_classification(
    language_label: str,
) -> list[ClassificationEntry] | None:
    match = process.extractOne(
        clean_label(language_label),
        cleaned_language_names,
        scorer=fuzz.WRatio,
        score_cutoff=90,
    )

    if not match:
        return None

    match_idx = match[2]
    matched_language = languages[match_idx]

    classification = [
        ClassificationEntry(
            classification_id=matched_language["language_id"],
            scientific_name=matched_language["name"],
            common_name=None,
            id_type=ClassificationEntryType.GLOTTOLOG_ID,
        )
    ]
    classification = extend_classification(classification)

    return classification


def extend_classification(
    classification: list[ClassificationEntry],
) -> list[ClassificationEntry]:
    highest_language = id_to_language[classification[-1].classification_id]
    next_parent_id = highest_language["parent_id"]
    next_parent = id_to_language.get(next_parent_id)

    if next_parent:
        classification.append(
            ClassificationEntry(
                classification_id=next_parent["language_id"],
                scientific_name=next_parent["name"],
                id_type=ClassificationEntryType.GLOTTOLOG_ID,
            )
        )
        classification = extend_classification(classification)

    return classification



from enum import Enum

import streamlit as st

from phylodata.data_types import (
    Metadata,
)
from phylodata.errors import ValidationError
from phylodata.process.evolutionary_model.evolutionary_model_module import (
    EvolutionaryModelModule,
)
from phylodata.process.experiment.experiment_module import ExperimentModule
from phylodata.process.files.files_module import FilesModule
from phylodata.process.paper.paper_module import PaperModule
from phylodata.process.samples.samples_module import SamplesModule
from phylodata.process.trees.trees_module import TreesModule
from phylodata.process.utils.generate_output import generate_output
from phylodata.version import __version__


class Stage(str, Enum):
    INPUT = "Input"
    NEXT_STEPS = "Next Steps"


STAGE = "stage"
OUTPUT_FOLDER = "output_folder"

paper_module = PaperModule()
experiment_module = ExperimentModule()
files_module = FilesModule()
evolutionary_model_module = EvolutionaryModelModule()
samples_module = SamplesModule()
trees_module = TreesModule()

if STAGE not in st.session_state:
    st.session_state[STAGE] = Stage.INPUT


if st.session_state[STAGE] == Stage.INPUT:
    """
    # 🌴 Process Experiment

    Let's process your Bayesian Phylogenetics experiment to prepare it for upload to the PhyloData repository.

    Note that right now, we only support BEAST 2 experiments.

    This usually takes less than five minutes.
    """

    paper_module.ui()
    experiment_module.ui()
    files_module.ui()

    """
    Almost done!
    """

    def process_input():
        with st.status("Processing your analysis...", expanded=True):
            try:
                paper_module.validate()
                experiment_module.validate()
                files_module.validate()
                experiment_module.validate()
                samples_module.validate()
                trees_module.validate()

                editable_paper, non_editable_paper = paper_module.parse()
                editable_experiment, non_editable_experiment = experiment_module.parse(
                    editable_paper
                )
                files = files_module.parse()
                parsed_evolutionary_model = evolutionary_model_module.parse(
                    files_module.beast2_configuration  # type: ignore
                )
                parsed_samples = samples_module.parse(files_module.beast2_configuration)  # type: ignore
                parsed_trees = trees_module.parse(files_module.beast2_trees)  # type: ignore
            except ValidationError as error:
                st.toast(error.message)
                return

            metadata = Metadata(evo_data_pipeline_version=__version__)

            st.text("Creating PhyloData folder...")

            output_folder = generate_output(
                editable_experiment,
                non_editable_experiment,
                editable_paper,
                non_editable_paper,
                parsed_samples,
                files,
                parsed_evolutionary_model,
                parsed_trees,
                metadata,
            )

            st.session_state[STAGE] = Stage.NEXT_STEPS
            st.session_state[OUTPUT_FOLDER] = output_folder
            st.rerun()

    process_experiment_button = st.button("Process experiment", type="primary")
    if process_experiment_button:
        process_input()


if st.session_state[STAGE] == Stage.NEXT_STEPS:
    """
    # 🚀 Next Steps

    You successfully processed your experiment! These are your next steps:

    1. Create a ZIP file of the newly created folder.
    2. Upload it on the [PhyloData website](http://phylodata.com/new-experiment).
    """



import taxoniq

from phylodata.data_types import ClassificationEntry, ClassificationEntryType


def look_up_taxon_classification(ncbi_taxon_id: int) -> list[ClassificationEntry]:
    """Returns the classification of the specified NCBI taxon."""
    taxon = taxoniq.Taxon(ncbi_taxon_id)

    classification = []
    for t in taxon.ranked_lineage:
        common_name = None
        try:
            common_name = t.common_name
        except taxoniq.NoValue:
            ...

        classification.append(
            ClassificationEntry(
                classification_id=str(t.tax_id),
                scientific_name=t.scientific_name,
                common_name=common_name,
                id_type=ClassificationEntryType.NCBI_TAXONOMY_ID,
            )
        )

    return classification



from io import BytesIO

from pytest import fixture

from phylodata.process.evolutionary_model.parse_beast2_packages import (
    BEAST2_PACKAGE_PARSERS,
    BEAST2_PACKAGE_PARSERS_PER_NAME,
    AutoGeneratedBeast2PackageParser,
)
from phylodata.process.evolutionary_model.parse_evolutionary_model import (
    parse_evolutionary_model,
)


def to_bytes_io(text: str):
    bytesio = BytesIO(bytes(text, "ascii"))
    bytesio.name = "Test"
    return bytesio


@fixture
def hky_model():
    package = AutoGeneratedBeast2PackageParser(
        "HKY", "treePrior", "beast.evolution.substitutionmodel.HKY"
    )
    BEAST2_PACKAGE_PARSERS.clear()
    BEAST2_PACKAGE_PARSERS.append(package)
    BEAST2_PACKAGE_PARSERS_PER_NAME["Babel"] = package


def test_detect_hky_with_no_namespace(hky_model):
    file = to_bytes_io("""<?xml version="1.0" encoding="UTF-8" standalone="no"?>
    <beast namespace="beast.core" required="BEAST.base v2.7.7" version="2.7">
        <data></data>
        <run>
            <distribution
                spec="ThreadedTreeLikelihood"
            >
                <substModel id="HKY" spec="beast.evolution.substitutionmodel.HKY" kappa="@kappa">
                    <frequencies id="Frequencies" spec="Frequencies" frequencies="@frequencies"/>
                </substModel>
            </distribution>
        </run>
    </beast>""")
    model = parse_evolutionary_model(file)

    assert model
    assert len(model) == 1
    assert model[0].name == "HKY"


def test_detect_hky_with_full_namespace():
    file = to_bytes_io("""<?xml version="1.0" encoding="UTF-8" standalone="no"?>
    <beast namespace="beast.core:beast.evolution.substitutionmodel" required="BEAST.base v2.7.7" version="2.7">
        <data></data>
        <run>
            <distribution
                spec="ThreadedTreeLikelihood"
            >
                <substModel id="HKY" spec="HKY" kappa="@kappa">
                    <frequencies id="Frequencies" spec="Frequencies" frequencies="@frequencies"/>
                </substModel>
            </distribution>
        </run>
    </beast>""")
    model = parse_evolutionary_model(file)

    assert model
    assert len(model) == 1
    assert model[0].name == "HKY"



"""This file contains all the types which are used to describe a PhyloData experiment.

For some types, there is an *editable* and *non-editable* version. We make this distinction
for the following reasons:

-   We can create two metadata files, making it very obvious which information can be manually
    corrected by the user and which information is an automated output of the pipeline.
-   Given the pipeline version, the non-editable information should always be reproducible.
-   When the pipeline changes (e.g. new features get added), the user-made corrections to
    the editable metadata can easily be integrated again.
"""

from datetime import date
from enum import Enum
from hashlib import md5
from io import BytesIO
from typing import Any, Optional

import msgspec


class ExperimentType(Enum):
    BEAST2_Experiment = "beast2Experiment"


class ClassificationEntryType(Enum):
    NCBI_TAXONOMY_ID = "ncbiTaxonomyId"
    GLOTTOLOG_ID = "glottologId"


class FileType(Enum):
    BEAST2_CONFIGURATION = "beast2Configuration"
    BEAST2_POSTERIOR_LOGS = "beast2PosteriorLogs"
    POSTERIOR_TREES = "posteriorTrees"
    SUMMARY_TREE = "summaryTree"
    CODEPHY_MODEL = "codephyModel"
    PHYLO_DATA_EXPERIMENT = "phyloDataExperiment"
    UNKNOWN = "unknown"


class DataType(Enum):
    RNA = "rna"
    DNA = "dna"
    AMINO_ACIDS = "aminoAcids"
    PHASED_DIPLOID_DNA = "phasedDiploidDna"
    TRAITS = "traits"
    UNKNOWN = "unknown"


class ModelType(Enum):
    SUBSTITUTION_MODEL = "substitutionModel"
    CLOCK_MODEL = "clockModel"
    TREE_PRIOR = "treePrior"
    TREE_LIKELIHOOD = "treeLikelihood"
    OTHER = "other"


class EditableExperiment(msgspec.Struct, rename="camel"):
    type: ExperimentType
    title: Optional[str] = None
    description: Optional[str] = None


class NonEditableExperiment(msgspec.Struct, rename="camel"):
    human_readable_id: str
    origin: str
    upload_date: date
    version: int
    license: str = "CC0"
    id: Optional[str] = None


class Experiment(msgspec.Struct, rename="camel"):
    type: ExperimentType
    human_readable_id: str
    origin: str
    upload_date: date
    version: int
    license: str = "CC0"
    title: Optional[str] = None
    description: Optional[str] = None
    id: Optional[str] = None

    @classmethod
    def from_partial(
        cls, editable: EditableExperiment, non_editable: NonEditableExperiment
    ):
        return cls(
            type=editable.type,
            human_readable_id=non_editable.human_readable_id,
            origin=non_editable.origin,
            upload_date=non_editable.upload_date,
            version=non_editable.version,
            license=non_editable.license,
            title=editable.title,
            description=editable.description,
            id=non_editable.id,
        )


class EditablePaper(msgspec.Struct, rename="camel"):
    title: str
    year: int
    authors: list[str]
    email: str
    abstract: str
    bibtex: str
    url: Optional[str] = None


class NonEditablePaper(msgspec.Struct, rename="camel"):
    doi: str


class Paper(msgspec.Struct, rename="camel"):
    doi: str
    title: str
    year: int
    authors: list[str]
    abstract: str
    bibtex: str
    url: Optional[str] = None

    @classmethod
    def from_partial(cls, editable: EditablePaper, non_editable: NonEditablePaper):
        return cls(
            doi=non_editable.doi,
            title=editable.title,
            year=editable.year,
            authors=editable.authors,
            abstract=editable.abstract,
            bibtex=editable.bibtex,
            url=editable.url,
        )


class File(
    msgspec.Struct, rename="camel", dict=True
):  # dict is true to omit private values (https://github.com/jcrist/msgspec/issues/199#issuecomment-2840826792)
    name: str
    type: FileType
    size_bytes: int
    md5: str
    is_preview: bool = False
    local_path: Optional[str] = None
    remote_path: Optional[str] = None

    def __post_init__(self):
        self.bytes: BytesIO | None = None

    def set_bytes(self, bytes: BytesIO) -> "File":
        self.bytes = bytes
        return self

    @classmethod
    def from_bytes(
        cls,
        bytes: BytesIO,
        name: str,
        type: FileType,
        is_preview: bool = False,
    ) -> "File":
        buffer = bytes.getbuffer()
        return File(
            name=name,
            type=type,
            is_preview=is_preview,
            size_bytes=buffer.nbytes,
            md5=md5(buffer).hexdigest(),
        ).set_bytes(bytes)


class SampleData(msgspec.Struct, rename="camel"):
    type: DataType
    length: int
    data: str


class ClassificationEntry(msgspec.Struct, rename="camel"):
    classification_id: str
    scientific_name: str
    id_type: ClassificationEntryType
    common_name: Optional[str] = None


class Sample(msgspec.Struct, rename="camel"):
    sample_id: str
    scientific_name: str
    classification: list[ClassificationEntry]
    sample_data: list[SampleData]
    common_name: Optional[str] = None


class Trees(msgspec.Struct, rename="camel"):
    number_of_trees: int
    number_of_tips: int
    ultrametric: bool
    time_tree: bool
    rooted: bool
    ccd1_entropy: float
    tree_ess: int
    ccd0_map_tree: str
    hipstr_tree: str
    leaf_to_sample_map: dict[str, str]
    average_root_age: float


class EvolutionaryModelComponent(msgspec.Struct, rename="camel"):
    name: str
    type: ModelType
    parameters: dict[str, Any]


class Metadata(msgspec.Struct, rename="camel"):
    evo_data_pipeline_version: str


class NonEditablePaperWithExperiment(msgspec.Struct, rename="camel"):
    """This structure contains all experiment data which is
    computed by the pipeline and must not be changed manually."""

    paper: NonEditablePaper
    experiment: NonEditableExperiment
    files: list[File]
    trees: Optional[Trees]
    evolutionary_model: list[EvolutionaryModelComponent]
    metadata: Metadata


class EditablePaperWithExperiment(msgspec.Struct, rename="camel"):
    """This structure contains all experiment data computed by the pipeline
    and may be changed manually."""

    paper: EditablePaper
    experiment: EditableExperiment
    samples: list[Sample]


class PaperWithExperiment(msgspec.Struct, rename="camel"):
    paper: Paper
    experiment: Experiment
    files: list[File]
    samples: list[Sample]
    trees: Optional[Trees]
    evolutionary_model: list[EvolutionaryModelComponent]
    metadata: Metadata

    @classmethod
    def from_partial(
        cls,
        editable: EditablePaperWithExperiment,
        non_editable: NonEditablePaperWithExperiment,
    ):
        return cls(
            paper=Paper.from_partial(editable.paper, non_editable.paper),
            experiment=Experiment.from_partial(
                editable.experiment, non_editable.experiment
            ),
            files=non_editable.files,
            samples=editable.samples,
            trees=non_editable.trees,
            evolutionary_model=non_editable.evolutionary_model,
            metadata=non_editable.metadata,
        )


def validate_editable_json(file_path: str):
    """Validates if the given file contains a valid EditablePaperWithExperiment object.
    Throws a msgspec.ValidationError if not."""
    with open(file_path) as handle:
        msgspec.json.decode(handle.read(), type=EditablePaperWithExperiment)


def get_schema():
    """Returns the formatted JSON schema corresponding to EditablePaperWithExperiment."""
    schema = msgspec.json.schema(EditablePaperWithExperiment)
    return msgspec.json.format(msgspec.json.encode(schema).decode())
